from git_code_debt.file_diff_stat import FileDiffStat
from git_code_debt.metrics.base import SimpleLineCounterBase


class Python__init__LineCount(SimpleLineCounterBase):
    def should_include_file(self, file_diff_stat: FileDiffStat) -> bool:
        return file_diff_stat.filename == b'__init__.py'

    def line_matches_metric(
            self,
            line: bytes,
            file_diff_stat: FileDiffStat,
    ) -> bool:
        # All lines in __init__.py match
        return True



from typing import NamedTuple
from typing import Set

from git_code_debt.server.presentation.delta import Delta


class CommitDelta(NamedTuple):
    metric_name: str
    classname: str
    delta: Delta

    @classmethod
    def from_data(
            cls,
            metric_name: str,
            delta: Delta,
            color_overrides: Set[str],
    ) -> 'CommitDelta':
        return cls(
            metric_name,
            # TODO: duplicated in Metric
            'color-override' if metric_name in color_overrides else '',
            delta,
        )



from typing import Generator
from typing import List
from typing import Tuple

from git_code_debt.file_diff_stat import FileDiffStat
from git_code_debt.file_diff_stat import SpecialFileType
from git_code_debt.file_diff_stat import Status
from git_code_debt.metric import Metric
from git_code_debt.metrics.base import DiffParserBase
from git_code_debt.metrics.base import MetricInfo
from git_code_debt.repo_parser import Commit


class BinaryFileCount(DiffParserBase):
    """Counts the number of _files_ considered to be binary by `git`."""

    def get_metrics_from_stat(
            self, _: Commit,
            file_diff_stats: Tuple[FileDiffStat, ...],
    ) -> Generator[Metric, None, None]:
        binary_delta = 0

        for file_diff_stat in file_diff_stats:
            if (
                    file_diff_stat.special_file is not None and (
                        file_diff_stat.special_file.file_type is
                        SpecialFileType.BINARY
                    )
            ):
                if file_diff_stat.status is Status.ADDED:
                    binary_delta += 1
                elif file_diff_stat.status is Status.DELETED:
                    binary_delta -= 1

        if binary_delta:
            yield Metric(type(self).__name__, binary_delta)

    def get_metrics_info(self) -> List[MetricInfo]:
        return [MetricInfo.from_class(type(self))]



import argparse
import os.path
import shutil
import sqlite3
from typing import Optional
from typing import Sequence
from typing import TYPE_CHECKING

import flask
import pkg_resources

from git_code_debt.server.metric_config import Config
from git_code_debt.server.servlets.changes import changes
from git_code_debt.server.servlets.commit import commit
from git_code_debt.server.servlets.graph import graph
from git_code_debt.server.servlets.index import index
from git_code_debt.server.servlets.status import status
from git_code_debt.server.servlets.widget import widget
from git_code_debt.util import yaml

if TYPE_CHECKING:
    from typing import NoReturn

app = flask.Flask(__name__)
app.register_blueprint(changes)
app.register_blueprint(commit)
app.register_blueprint(graph)
app.register_blueprint(index)
app.register_blueprint(status)
app.register_blueprint(widget)


class AppContext:
    database_path = 'database.db'
    config: Optional[Config] = None


@app.before_request
def before_request() -> None:
    flask.g.config = AppContext.config
    flask.g.db = sqlite3.connect(AppContext.database_path)


@app.teardown_request
def teardown_request(_: Optional[Exception]) -> None:
    flask.g.config = None
    flask.g.db.close()


def create_metric_config_if_not_exists() -> None:
    """Creates the sameple metric_config.yaml in the current directory if
    it does not exist.
    """
    if os.path.exists('metric_config.yaml'):
        return

    print('WARNING: no metric_config.yaml detected.  Creating sample config!')
    shutil.copyfile(
        pkg_resources.resource_filename(
            'git_code_debt.server', 'metric_config.sample.yaml',
        ),
        'metric_config.yaml',
    )


def main(argv: Optional[Sequence[str]] = None) -> 'NoReturn':
    parser = argparse.ArgumentParser()
    parser.add_argument('database_path', type=str)
    parser.add_argument('-p', '--port', type=int, default=5000)
    mutex = parser.add_mutually_exclusive_group()
    mutex.add_argument(
        '--debug', action='store_true',
        help=(
            'Run in debug mode (stacktraces + single process). '
            'Not suggested for production.'
        ),
    )
    mutex.add_argument(
        '--processes', type=int, default=5,
        help='Number of processes, does not apply to --debug',
    )
    args = parser.parse_args(argv)

    if not os.path.exists(args.database_path):
        print(f'Not found: {args.database_path}')
        print('Use git-code-debt-generate to create a database.')
        raise SystemExit(1)

    create_metric_config_if_not_exists()
    with open('metric_config.yaml') as f:
        contents = yaml.load(f)
    AppContext.config = Config.from_data(contents)

    AppContext.database_path = args.database_path
    kwargs = {'port': args.port, 'debug': args.debug}
    if not args.debug:
        kwargs['processes'] = args.processes
        kwargs['threaded'] = False

    app.run('0.0.0.0', **kwargs)
    raise SystemExit(1)


if __name__ == '__main__':
    exit(main())



from types import ModuleType
from typing import Any
from typing import List
from typing import Sequence
from typing import Set
from typing import Type

from git_code_debt.metrics.base import DiffParserBase
from git_code_debt.util.discovery import discover


def is_metric_cls(cls: Type[Any]) -> bool:
    """A metric class is defined as follows:

        - It inherits DiffParserBase
        - It is not DiffParserBase
        - It does not have __metric__ = False
    """
    return (
        cls is not DiffParserBase and
        cls.__dict__.get('__metric__', True) and
        issubclass(cls, DiffParserBase)
    )


def get_metric_parsers(
        metric_packages: Sequence[ModuleType] = (),
        include_defaults: bool = True,
) -> Set[Type[DiffParserBase]]:
    """Gets all of the metric parsers.

    Args:
        metric_packages - Defaults to no extra packages. An iterable of
            metric containing packages.  A metric inherits DiffParserBase
            and does not have __metric__ = False
            A metric package must be imported using import a.b.c
        include_defaults - Whether to include the generic metric parsers
    """
    metric_parsers = set()

    if include_defaults:
        import git_code_debt.metrics
        metric_parsers.update(discover(git_code_debt.metrics, is_metric_cls))

    for metric_package in metric_packages:
        metric_parsers.update(discover(metric_package, is_metric_cls))
    return metric_parsers


def get_modules(module_names: List[str]) -> List[ModuleType]:
    """Returns module objects for each module name.  Has the side effect of
    importing each module.

    Args:
        module_names - iterable of module names

    Returns:
        Module objects for each module specified in module_names
    """
    return [
        __import__(module_name, fromlist=['__trash__'])
        for module_name in module_names
    ]


def get_metric_parsers_from_args(
        metric_package_names: List[str],
        skip_defaults: bool,
) -> Set[Type[DiffParserBase]]:
    packages = get_modules(metric_package_names)
    return get_metric_parsers(
        metric_packages=packages, include_defaults=not skip_defaults,
    )



from identify import identify

UNKNOWN = 'unknown'
IGNORED_TAGS = frozenset((
    identify.DIRECTORY, identify.SYMLINK, identify.FILE,
    identify.EXECUTABLE, identify.NON_EXECUTABLE,
    identify.TEXT, identify.BINARY,
))
ALL_TAGS = frozenset((identify.ALL_TAGS - IGNORED_TAGS) | {UNKNOWN})



import collections
from typing import Dict
from typing import Generator
from typing import List
from typing import Tuple

from identify import identify

from git_code_debt.file_diff_stat import FileDiffStat
from git_code_debt.metric import Metric
from git_code_debt.metrics.base import DiffParserBase
from git_code_debt.metrics.base import MetricInfo
from git_code_debt.metrics.common import ALL_TAGS
from git_code_debt.metrics.common import UNKNOWN
from git_code_debt.metrics.curse_words import word_list
from git_code_debt.repo_parser import Commit


def count_curse_words(lines: List[bytes]) -> int:
    curses = 0
    for line in lines:
        for word in line.split():
            if word in word_list:
                curses += 1
    return curses


class CurseWordsParser(DiffParserBase):
    """Counts curse words in a repository, overall and by file type"""

    def get_metrics_from_stat(
            self,
            _: Commit,
            file_diff_stats: Tuple[FileDiffStat, ...],
    ) -> Generator[Metric, None, None]:
        total_curses = 0
        curses_by_file_type: Dict[str, int] = collections.defaultdict(int)

        for file_diff_stat in file_diff_stats:
            curses_added = count_curse_words(file_diff_stat.lines_added)
            curses_removed = count_curse_words(file_diff_stat.lines_removed)
            curses_changed = curses_added - curses_removed

            # Track total overall
            total_curses = total_curses + curses_changed

            # Track by file extension -> type mapping
            filename = file_diff_stat.filename.decode('UTF-8')
            tags = identify.tags_from_filename(filename) or {UNKNOWN}

            for tag in tags:
                curses_by_file_type[tag] += curses_changed

        # Yield overall metric and one per type of expected mapping types
        yield Metric('TotalCurseWords', total_curses)
        for tag, value in curses_by_file_type.items():
            if tag in ALL_TAGS and value:
                yield Metric(f'TotalCurseWords_{tag}', value)

    def get_metrics_info(self) -> List[MetricInfo]:
        metric_names = [f'TotalCurseWords_{tag}' for tag in ALL_TAGS]
        metric_names.append('TotalCurseWords')
        return [MetricInfo(metric_name) for metric_name in metric_names]



import functools

import yaml


Loader = getattr(yaml, 'CSafeLoader', yaml.SafeLoader)
load = functools.partial(yaml.load, Loader=Loader)
Dumper = getattr(yaml, 'CSafeDumper', yaml.SafeDumper)
dump = functools.partial(yaml.dump, Dumper=Dumper)



import argparse
from typing import Optional
from typing import Sequence

from git_code_debt import options
from git_code_debt.discovery import get_metric_parsers_from_args
from git_code_debt.generate import get_options_from_config


CYAN = '\033[1;36m'
NORMAL = '\033[0m'


def color(text: str, color_value: str, color_setting: bool) -> str:
    if not color_setting:
        return text
    else:
        return f'{color_value}{text}{NORMAL}'


def main(argv: Optional[Sequence[str]] = None) -> int:
    parser = argparse.ArgumentParser(description='List metric parsers')
    # optional
    options.add_color(parser)
    options.add_generate_config_filename(parser)
    parsed_args = parser.parse_args(argv)

    color_setting = parsed_args.color in ('always', 'auto')
    args = get_options_from_config(parsed_args.config_filename)

    metric_parsers = get_metric_parsers_from_args(
        args.metric_package_names,
        args.skip_default_metrics,
    )

    metric_parsers_sorted = sorted(
        metric_parsers,
        key=lambda cls: cls.__module__ + cls.__name__,
    )

    for metric_parser_cls in metric_parsers_sorted:
        print(
            '{} {}'.format(
                color(metric_parser_cls.__module__, CYAN, color_setting),
                metric_parser_cls.__name__,
            ),
        )
        for name, description in metric_parser_cls().get_metrics_info():
            description = f': {description}' if description else ''
            print(f'    {name}{description}')
    return 0


if __name__ == '__main__':
    exit(main())



import sqlite3
from typing import Dict
from typing import Tuple

from git_code_debt.metric import Metric
from git_code_debt.repo_parser import Commit


def insert_metric_values(
        db: sqlite3.Connection,
        metric_values: Dict[int, int],
        has_data: Dict[int, bool],
        commit: Commit,
) -> None:
    values = [
        (commit.sha, metric_id, commit.date, value)
        for metric_id, value in metric_values.items()
        if has_data[metric_id]
    ]
    db.executemany(
        'INSERT INTO metric_data (sha, metric_id, timestamp, running_value)\n'
        'VALUES (?, ?, ?, ?)\n',
        values,
    )


def insert_metric_changes(
        db: sqlite3.Connection,
        metrics: Tuple[Metric, ...],
        metric_mapping: Dict[str, int],
        commit: Commit,
) -> None:
    """Insert into the metric_changes tables.

    :param metrics: `list` of `Metric` objects
    :param dict metric_mapping: Maps metric names to ids
    :param Commit commit:
    """
    values = [
        [commit.sha, metric_mapping[metric.name], metric.value]
        for metric in metrics
        if metric.value != 0
    ]
    db.executemany(
        'INSERT INTO metric_changes (sha, metric_id, value) VALUES (?, ?, ?)',
        values,
    )



import inspect
from typing import Generator
from typing import List
from typing import NamedTuple
from typing import Tuple
from typing import Type

from git_code_debt.file_diff_stat import FileDiffStat
from git_code_debt.metric import Metric
from git_code_debt.repo_parser import Commit


class MetricInfo(NamedTuple):
    name: str
    description: str = ''

    @classmethod
    def from_class(cls, c: 'Type[DiffParserBase]') -> 'MetricInfo':
        return cls(c.__name__, inspect.cleandoc(c.__doc__ or ''))


class DiffParserBase:
    # Specify __metric__ = False to not be included (useful for base classes)
    __metric__ = False

    def get_metrics_from_stat(
            self,
            commit: Commit,
            file_diff_stats: Tuple[FileDiffStat, ...],
    ) -> Generator[Metric, None, None]:
        """Implement me to yield Metric objects from the input list of
        FileStat objects.

        Args:
            commit - Commit object
            file_diff_stats - list of FileDiffStat objects

        Returns:
           generator of Metric objects
        """
        raise NotImplementedError

    def get_metrics_info(self) -> List[MetricInfo]:
        raise NotImplementedError


class SimpleLineCounterBase(DiffParserBase):
    __metric__ = False

    def get_metrics_from_stat(
            self,
            _: Commit,
            file_diff_stats: Tuple[FileDiffStat, ...],
    ) -> Generator[Metric, None, None]:
        metric_value = 0

        for file_diff_stat in file_diff_stats:
            if self.should_include_file(file_diff_stat):
                for line in file_diff_stat.lines_added:
                    if self.line_matches_metric(line, file_diff_stat):
                        metric_value += 1
                for line in file_diff_stat.lines_removed:
                    if self.line_matches_metric(line, file_diff_stat):
                        metric_value -= 1

        if metric_value:
            yield Metric(self.metric_name, metric_value)

    def get_metrics_info(self) -> List[MetricInfo]:
        return [MetricInfo(self.metric_name, self.metric_description)]

    @property
    def metric_name(self) -> str:
        """Override me or make a class-level metric_name attribute to set the
        metric name.  Defaults to class name
        """
        return type(self).__name__

    @property
    def metric_description(self) -> str:
        """Override me or make a class-level metric_description attribute to
        set the metric description.  Defaults to metric docstring.
        """
        return inspect.cleandoc(type(self).__doc__ or '')

    def should_include_file(self, file_diff_stat: FileDiffStat) -> bool:
        """Implement me to return whether a filename should be included.
        By default, this returns True.

        :param FileDiffStat file_diff_stat:
        """
        return True

    def line_matches_metric(
            self,
            line: bytes,
            file_diff_stat: FileDiffStat,
    ) -> bool:
        """Implement me to return whether a line matches the metric.

        :param bytes line: Line in the file
        :param FileDiffStat file_diff_stat:
        """
        raise NotImplementedError


class TextLineCounterBase(SimpleLineCounterBase):
    __metric__ = False

    def text_line_matches_metric(
            self,
            line: str,
            file_diff_stat: FileDiffStat,
    ) -> bool:
        raise NotImplementedError

    def line_matches_metric(
            self,
            line: bytes,
            file_diff_stat: FileDiffStat,
    ) -> bool:
        text = line.decode('UTF-8', 'ignore')
        return self.text_line_matches_metric(text, file_diff_stat)



import enum
import os.path
import re
from typing import List
from typing import NamedTuple
from typing import Optional


Status = enum.Enum('Status', 'ADDED DELETED ALREADY_EXISTING')
SpecialFileType = enum.Enum('SpecialFileType', 'SUBMODULE SYMLINK BINARY')


class SpecialFile(NamedTuple):
    file_type: SpecialFileType
    added: Optional[bytes]
    removed: Optional[bytes]


class FileDiffStat(NamedTuple):
    path: bytes
    lines_added: List[bytes]
    lines_removed: List[bytes]
    status: Status
    special_file: Optional[SpecialFile] = None

    @property
    def extension(self) -> bytes:
        return os.path.splitext(self.path)[1]

    @property
    def filename(self) -> bytes:
        return os.path.split(self.path)[1]


SUBMODULE_MODE = b'160000'
SYMLINK_MODE = b'120000'


def _to_file_diff_stat(file_diff: bytes) -> FileDiffStat:
    lines = file_diff.split(b'\n')
    diff_line_filename = lines[0].split()[-1].lstrip(b'b').lstrip(b'/')
    is_binary = False
    in_diff = False
    mode = None
    status = None
    lines_added = []
    lines_removed = []

    for line in lines[1:]:
        # Mode will be indicated somewhere between diff --git line
        # and the file added / removed lines
        # It has these forms:
        # 1. 'new file mode 100644'
        # 2. 'deleted file mode 100644'
        # 3. 'index dc7827c..7b8b995 100644'
        # 4. 'old mode 100755'
        #    'new mode 100644'
        if line.startswith(b'new file mode '):
            assert status is None
            assert mode is None
            status = Status.ADDED
            mode = line.split()[-1]
        elif line.startswith(b'deleted file mode '):
            assert status is None
            assert mode is None
            status = Status.DELETED
            mode = line.split()[-1]
        elif line.startswith(b'new mode '):
            assert status is None
            assert mode is None
            status = Status.ALREADY_EXISTING
            mode = line.split()[-1]
        elif line.startswith(b'index') and len(line.split()) == 3:
            assert status is None
            assert mode is None
            status = Status.ALREADY_EXISTING
            mode = line.split()[-1]
        elif line.startswith(b'Binary files'):
            is_binary = True
        # A diff contains lines that look like:
        # --- foo/bar
        # +++ foo/bar
        # Which kind of look like diff lines but are definitely not
        elif line.startswith(b'--- ') and not in_diff:
            pass
        elif line.startswith(b'+++ ') and not in_diff:
            in_diff = True
        elif in_diff and line.startswith(b'+'):
            lines_added.append(line[1:])
        elif in_diff and line.startswith(b'-'):
            lines_removed.append(line[1:])

    assert mode is not None
    assert status is not None

    # Process symlinks and submodules
    special_file = None
    if mode == SUBMODULE_MODE:
        special_file = SpecialFile(
            file_type=SpecialFileType.SUBMODULE,
            added=lines_added[0].split()[-1] if lines_added else None,
            removed=lines_removed[0].split()[-1] if lines_removed else None,
        )
        lines_added = []
        lines_removed = []
    elif mode == SYMLINK_MODE:
        special_file = SpecialFile(
            file_type=SpecialFileType.SYMLINK,
            added=lines_added[0] if lines_added else None,
            removed=lines_removed[0] if lines_removed else None,
        )
        lines_added = []
        lines_removed = []
    elif is_binary:
        special_file = SpecialFile(
            file_type=SpecialFileType.BINARY,
            added=diff_line_filename if status is not Status.DELETED else None,
            removed=diff_line_filename if status is not Status.ADDED else None,
        )

    return FileDiffStat(
        diff_line_filename,
        lines_added,
        lines_removed,
        status,
        special_file=special_file,
    )


GIT_DIFF_RE = re.compile(b'^diff --git', flags=re.MULTILINE)


def get_file_diff_stats_from_output(output: bytes) -> List[FileDiffStat]:
    files = GIT_DIFF_RE.split(output)
    assert not files[0].strip() or files[0].startswith(b'commit ')
    return [_to_file_diff_stat(file_diff) for file_diff in files[1:]]



import calendar
import datetime
from typing import Tuple


def to_timestamp(dt: datetime.datetime) -> int:
    return calendar.timegm(dt.utctimetuple())


def data_points_for_time_range(
        start_timestamp: int,
        end_timestamp: int,
        data_points: int = 25,
) -> Tuple[int, ...]:
    interval = ((end_timestamp - start_timestamp) // data_points) or 1
    return tuple(range(start_timestamp, end_timestamp + interval, interval))



from git_code_debt.file_diff_stat import FileDiffStat
from git_code_debt.metrics.base import SimpleLineCounterBase


def is_python_import(line: bytes) -> bool:
    line = line.lstrip()
    if line.startswith(b'import'):
        return True
    if line.startswith(b'from') and b'import' in line:
        return True
    return False


def is_template_import(line: bytes) -> bool:
    line = line.lstrip()
    return (
        line.startswith(b'#') and
        is_python_import(line[1:])
    )


class PythonImportCount(SimpleLineCounterBase):
    def should_include_file(self, file_diff_stat: FileDiffStat) -> bool:
        return file_diff_stat.extension == b'.py'

    def line_matches_metric(
            self,
            line: bytes,
            file_diff_stat: FileDiffStat,
    ) -> bool:
        return is_python_import(line)


class CheetahTemplateImportCount(SimpleLineCounterBase):
    def should_include_file(self, file_diff_stat: FileDiffStat) -> bool:
        return file_diff_stat.extension == b'.tmpl'

    def line_matches_metric(
            self,
            line: bytes,
            file_diff_stat: FileDiffStat,
    ) -> bool:
        return is_template_import(line)



import json

import flask

from git_code_debt.discovery import get_metric_parsers_from_args
from git_code_debt.generate import get_metrics
from git_code_debt.generate_config import GenerateOptions
from git_code_debt.repo_parser import BLANK_COMMIT
from git_code_debt.server.presentation.commit_delta import CommitDelta
from git_code_debt.server.presentation.delta import Delta
from git_code_debt.server.render_mako import render_template
from git_code_debt.util import yaml


widget = flask.Blueprint('widget', __name__)


@widget.route('/widget/frame')
def frame() -> str:
    return render_template('widget_frame.mako')


@widget.route('/widget/data', methods=['POST'])
def data() -> str:
    metric_names = frozenset(flask.g.config.widget_metrics)
    diff = flask.request.form['diff'].encode('UTF-8')

    metric_config = GenerateOptions.from_yaml(
        yaml.load(open('generate_config.yaml').read()),
    )
    parsers = get_metric_parsers_from_args(
        metric_config.metric_package_names, skip_defaults=False,
    )
    metrics = get_metrics(BLANK_COMMIT, diff, parsers, metric_config.exclude)
    metrics = tuple(
        metric for metric in metrics
        if metric.value and metric.name in metric_names
    )

    commit_deltas = sorted(
        CommitDelta.from_data(
            metric.name, Delta('javascript:;', metric.value),
            color_overrides=flask.g.config.color_overrides,
        )
        for metric in metrics
    )
    return json.dumps({
        'metrics': render_template('widget.mako', commit_deltas=commit_deltas),
    })



import argparse
import collections
import contextlib
import itertools
import multiprocessing.pool
import os.path
import sqlite3
from typing import Callable
from typing import Counter
from typing import Dict
from typing import Generator
from typing import Iterable
from typing import List
from typing import Optional
from typing import Pattern
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Type
from typing import TypeVar

import pkg_resources

from git_code_debt import options
from git_code_debt.discovery import get_metric_parsers_from_args
from git_code_debt.file_diff_stat import FileDiffStat
from git_code_debt.file_diff_stat import get_file_diff_stats_from_output
from git_code_debt.generate_config import GenerateOptions
from git_code_debt.logic import get_metric_has_data
from git_code_debt.logic import get_metric_mapping
from git_code_debt.logic import get_metric_values
from git_code_debt.logic import get_previous_sha
from git_code_debt.metric import Metric
from git_code_debt.metrics.base import DiffParserBase
from git_code_debt.metrics.base import MetricInfo
from git_code_debt.repo_parser import Commit
from git_code_debt.repo_parser import RepoParser
from git_code_debt.util import yaml
from git_code_debt.write_logic import insert_metric_changes
from git_code_debt.write_logic import insert_metric_values


def get_metrics(
        commit: Commit,
        diff: bytes, metric_parsers:
        Set[Type[DiffParserBase]],
        exclude: Pattern[bytes],
) -> Tuple[Metric, ...]:
    def get_all_metrics(
            file_diff_stats: Tuple[FileDiffStat, ...],
    ) -> Generator[Metric, None, None]:
        for metric_parser_cls in metric_parsers:
            metric_parser = metric_parser_cls()
            yield from metric_parser.get_metrics_from_stat(
                commit, file_diff_stats,
            )

    file_diff_stats = tuple(
        x for x in get_file_diff_stats_from_output(diff)
        if not exclude.search(x.path)
    )
    return tuple(get_all_metrics(file_diff_stats))


def increment_metrics(
        metric_values: Counter[int],
        metric_mapping: Dict[str, int],
        metrics: Tuple[Metric, ...],
) -> None:
    metric_values.update({metric_mapping[m.name]: m.value for m in metrics})


def _get_metrics_inner(
        mp_args: Tuple[
            Optional[Commit],
            Commit,
            RepoParser,
            Set[Type[DiffParserBase]],
            Pattern[bytes],
        ],
) -> Tuple[Metric, ...]:
    compare_commit, commit, repo_parser, metric_parsers, exclude = mp_args
    if compare_commit is None:
        diff = repo_parser.get_original_commit(commit.sha)
    else:
        diff = repo_parser.get_commit_diff(compare_commit.sha, commit.sha)
    return get_metrics(commit, diff, metric_parsers, exclude)


T = TypeVar('T')
T2 = TypeVar('T2')


@contextlib.contextmanager
def mapper(jobs: int) -> Generator[
    Callable[[Callable[[T2], T], Iterable[T2]], Iterable[T]], None, None,
]:
    if jobs == 1:
        yield map
    else:
        with contextlib.closing(multiprocessing.Pool(jobs)) as pool:
            yield pool.imap


def update_has_data(
        db: sqlite3.Connection,
        metrics: Tuple[Metric, ...],
        metric_mapping: Dict[str, int],
        has_data: Dict[int, bool],
) -> None:
    query = 'UPDATE metric_names SET has_data=1 WHERE id = ?'
    for metric_id in [metric_mapping[m.name] for m in metrics if m.value]:
        if not has_data[metric_id]:
            has_data[metric_id] = True
            db.execute(query, (metric_id,))


def load_data(
        database_file: str,
        repo: str,
        package_names: List[str],
        skip_defaults: bool,
        exclude: Pattern[bytes],
        jobs: int,
) -> None:
    metric_parsers = get_metric_parsers_from_args(package_names, skip_defaults)

    with sqlite3.connect(database_file) as db:
        metric_mapping = get_metric_mapping(db)
        has_data = get_metric_has_data(db)

        repo_parser = RepoParser(repo)

        with repo_parser.repo_checked_out():
            previous_sha = get_previous_sha(db)
            commits = repo_parser.get_commits(since_sha=previous_sha)

            # If there is nothing to check gtfo
            if len(commits) == 1 and previous_sha is not None:
                return

            # Maps metric_id to a running value
            metric_values: Counter[int] = collections.Counter()

            # Grab the state of our metrics at the last place
            compare_commit = None
            if previous_sha is not None:
                compare_commit = commits.pop(0)
                metric_values.update(get_metric_values(db, compare_commit.sha))

            mp_args = zip(
                [compare_commit, *commits],
                commits,
                itertools.repeat(repo_parser),
                itertools.repeat(metric_parsers),
                itertools.repeat(exclude),
            )
            with mapper(jobs) as do_map:
                for commit, metrics in zip(
                        commits, do_map(_get_metrics_inner, mp_args),
                ):
                    update_has_data(db, metrics, metric_mapping, has_data)
                    increment_metrics(metric_values, metric_mapping, metrics)
                    insert_metric_values(db, metric_values, has_data, commit)
                    insert_metric_changes(db, metrics, metric_mapping, commit)


def create_schema(db: sqlite3.Connection) -> None:
    """Creates the database schema."""
    schema_dir = pkg_resources.resource_filename('git_code_debt', 'schema')
    schema_files = os.listdir(schema_dir)

    for sql_file in schema_files:
        resource_filename = os.path.join(schema_dir, sql_file)
        with open(resource_filename) as resource:
            db.executescript(resource.read())


def get_metrics_info(
        metric_parsers: Set[Type[DiffParserBase]],
) -> List[MetricInfo]:
    metrics_info = set()
    for metric_parser_cls in metric_parsers:
        metrics_info.update(metric_parser_cls().get_metrics_info())
    return sorted(metrics_info)


def insert_metrics_info(
        db: sqlite3.Connection,
        metrics_info: List[MetricInfo],
) -> None:
    query = 'INSERT INTO metric_names (name, description) VALUES (?, ?)'
    db.executemany(query, metrics_info)


def populate_metric_ids(
        db: sqlite3.Connection,
        package_names: List[str],
        skip_defaults: bool,
) -> None:
    metric_parsers = get_metric_parsers_from_args(package_names, skip_defaults)
    metrics_info = get_metrics_info(metric_parsers)
    insert_metrics_info(db, metrics_info)


def create_database(args: GenerateOptions) -> None:
    with sqlite3.connect(args.database) as db:
        create_schema(db)
        populate_metric_ids(
            db,
            args.metric_package_names,
            args.skip_default_metrics,
        )


def get_options_from_config(config_filename: str) -> GenerateOptions:
    if not os.path.exists(config_filename):
        print(f'config file not found {config_filename}')
        exit(1)

    with open(config_filename) as config_file:
        return GenerateOptions.from_yaml(yaml.load(config_file))


def main(argv: Optional[Sequence[str]] = None) -> int:
    parser = argparse.ArgumentParser()
    options.add_generate_config_filename(parser)
    parser.add_argument(
        '-j', '--jobs', type=int, default=multiprocessing.cpu_count(),
    )
    parsed_args = parser.parse_args(argv)
    args = get_options_from_config(parsed_args.config_filename)

    if not os.path.exists(args.database):
        create_database(args)

    load_data(
        args.database,
        args.repo,
        args.metric_package_names,
        args.skip_default_metrics,
        args.exclude,
        parsed_args.jobs,
    )
    return 0


if __name__ == '__main__':
    exit(main())



import sqlite3
from typing import Dict
from typing import Optional


def get_metric_mapping(db: sqlite3.Connection) -> Dict[str, int]:
    """Gets a mapping from metric_name to metric_id."""
    results = db.execute('SELECT name, id FROM metric_names').fetchall()
    return dict(results)


def get_metric_has_data(db: sqlite3.Connection) -> Dict[int, bool]:
    res = db.execute('SELECT id, has_data FROM metric_names').fetchall()
    return {k: bool(v) for k, v in res}


def get_previous_sha(db: sqlite3.Connection) -> Optional[str]:
    """Gets the latest inserted SHA."""
    result = db.execute(
        # Use ROWID as a free, auto-incrementing, primary key.
        'SELECT sha FROM metric_data ORDER BY ROWID DESC LIMIT 1',
    ).fetchone()

    return result[0] if result else None


def get_metric_values(db: sqlite3.Connection, sha: str) -> Dict[int, int]:
    """Gets the metric values from a specific commit.

    :param db: Database object
    :param text sha: A sha representing a single commit
    """
    results = db.execute(
        'SELECT metric_id, running_value FROM metric_data WHERE sha = ?',
        (sha,),
    )
    return dict(results)



import datetime
import json

import flask

from git_code_debt.server import logic
from git_code_debt.server.presentation.commit_delta import CommitDelta
from git_code_debt.server.presentation.delta import Delta
from git_code_debt.server.render_mako import render_template


changes = flask.Blueprint('changes', __name__)


@changes.route('/changes/<metric_name>/<int:start_ts>/<int:end_ts>')
def show(metric_name: str, start_ts: int, end_ts: int) -> str:
    metric_info = logic.get_metric_info(flask.g.db, metric_name)

    metric_changes_from_db = sorted(
        logic.get_major_changes_for_metric(
            flask.g.db, start_ts, end_ts, metric_info.id,
        ),
    )
    metric_changes = [
        (
            datetime.datetime.fromtimestamp(timestamp).strftime(
                '%Y-%m-%d %H:%M:%S',
            ),
            sha,
            CommitDelta.from_data(
                metric_name, Delta('javascript:;', value),
                color_overrides=flask.g.config.color_overrides,
            ),
        )
        for timestamp, sha, value in metric_changes_from_db
    ]

    override_classname = (
        'color-override'
        if metric_name in flask.g.config.color_overrides
        else ''
    )

    rendered_template = render_template(
        'changes.mako',
        changes=metric_changes,
        override_classname=override_classname,
    )

    return json.dumps({'body': rendered_template})



import collections
import sqlite3
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple

import flask


class Metric(NamedTuple):
    value: int
    date: int


class MetricInfo(NamedTuple):
    id: int
    description: str


def get_metric_ids(db: sqlite3.Connection) -> List[str]:
    query = 'SELECT name FROM metric_names WHERE has_data=1 ORDER BY name'
    res = db.execute(query).fetchall()
    return [name for name, in res]


def get_metric_info(db: sqlite3.Connection, metric_name: str) -> MetricInfo:
    query = 'SELECT id, description FROM metric_names WHERE name = ?'
    res = db.execute(query, (metric_name,)).fetchone()
    return MetricInfo(*res)


def get_latest_sha() -> Optional[str]:
    query = 'SELECT sha FROM metric_data ORDER BY timestamp DESC LIMIT 1'
    result = flask.g.db.execute(query).fetchone()

    # If there is no data result will be None
    return result[0] if result else None


def get_sha_for_date(date: int) -> Optional[str]:
    result = flask.g.db.execute(
        '\n'.join((
            'SELECT',
            '    sha',
            'FROM metric_data',
            'WHERE',
            '    timestamp <= ?',
            'ORDER BY timestamp DESC',
            'LIMIT 1',
        )),
        [date],
    ).fetchone()

    # If the date is too far in the past (before data) there won't be a result
    return result[0] if result else None


def get_metrics_for_sha(sha: Optional[str]) -> Dict[str, int]:
    # For no sha, we default all metrics to 0
    if not sha:
        return collections.defaultdict(int)

    result = flask.g.db.execute(
        'SELECT\n'
        '    metric_names.name,\n'
        '    metric_data.running_value\n'
        'FROM metric_data\n'
        'INNER JOIN metric_names ON\n'
        '    metric_names.id = metric_data.metric_id AND\n'
        '    metric_names.has_data = 1\n'
        'WHERE\n'
        '    metric_data.sha = ?\n',
        [sha],
    ).fetchall()

    return collections.defaultdict(int, result)


def metrics_for_dates(metric_id: int, dates: Tuple[int, ...]) -> List[Metric]:
    def get_metric_for_timestamp(timestamp: int) -> Metric:
        result = flask.g.db.execute(
            'SELECT running_value, timestamp\n'
            'FROM metric_data\n'
            'WHERE metric_id = ? AND timestamp < ?\n'
            'ORDER BY timestamp DESC\n'
            'LIMIT 1\n',
            (metric_id, timestamp),
        ).fetchone()
        if result:
            return Metric(*result)
        else:
            return Metric(0, timestamp)

    return [get_metric_for_timestamp(date) for date in dates]


def get_first_data_timestamp(
        metric_name: str,
        db: Optional[sqlite3.Connection] = None,
) -> int:
    db = db or flask.g.db

    # Find the first change for that metric
    first_timestamp = db.execute(
        'SELECT timestamp\n'
        'FROM metric_data\n'
        'INNER JOIN metric_names ON metric_names.id = metric_data.metric_id\n'
        'WHERE metric_names.name = ?\n'
        'ORDER BY metric_data.ROWID ASC\n'
        'LIMIT 1\n',
        (metric_name,),
    ).fetchone()
    if not first_timestamp:
        return 0
    else:
        return first_timestamp[0]


def get_metric_changes(
        db: sqlite3.Connection,
        sha: str,
) -> List[Tuple[str, int]]:
    return db.execute(
        '\n'.join((
            'SELECT',
            '    metric_names.name,',
            '    metric_changes.value',
            'FROM metric_changes',
            'INNER JOIN metric_names',
            '    ON metric_changes.metric_id = metric_names.id',
            'WHERE metric_changes.sha = ?',
        )),
        [sha],
    ).fetchall()


def get_major_changes_for_metric(
        db: sqlite3.Connection,
        start_timestamp: int,
        end_timestamp: int,
        metric_id: int,
) -> List[Tuple[int, str, int]]:
    return db.execute(
        '\n'.join((
            'SELECT',
            '    metric_data.timestamp,',
            '    metric_data.sha,',
            '    metric_changes.value',
            'FROM metric_changes',
            'INNER JOIN metric_data ON',
            '    metric_changes.sha = metric_data.sha AND',
            '    metric_changes.metric_id = metric_data.metric_id',
            'WHERE',
            '    metric_data.timestamp >= ? AND',
            '    metric_data.timestamp < ? AND',
            '    metric_changes.metric_id = ?',
            'ORDER BY ABS(metric_changes.value) DESC',
            'LIMIT 50',
        )),
        (start_timestamp, end_timestamp, metric_id),
    ).fetchall()



from typing import NamedTuple


SIGNS_TO_CLASSNAMES = {
    0: 'metric-none',
    1: 'metric-up',
    -1: 'metric-down',
}


class Delta(NamedTuple):
    url: str
    value: int

    @property
    def classname(self) -> str:
        if not self.value:
            sign = 0
        else:
            sign = self.value // abs(self.value)

        return SIGNS_TO_CLASSNAMES[sign]



import flask

from git_code_debt.server import logic
from git_code_debt.server.presentation.commit_delta import CommitDelta
from git_code_debt.server.presentation.delta import Delta
from git_code_debt.server.render_mako import render_template


commit = flask.Blueprint('commit', __name__)


@commit.route('/commit/<sha>')
def show(sha: str) -> str:
    changes = logic.get_metric_changes(flask.g.db, sha)

    commit_deltas = sorted(
        CommitDelta.from_data(
            metric_name, Delta('javascript:;', change),
            color_overrides=flask.g.config.color_overrides,
        )
        for metric_name, change in changes
    )

    links = [
        (link_name, link.format(sha=sha))
        for link_name, link in flask.g.config.commit_links
    ]

    return render_template(
        'commit.mako',
        sha=sha,
        short_sha=sha[:8],
        commit_deltas=commit_deltas,
        links=links,
    )



import subprocess
from typing import Any


class CalledProcessError(RuntimeError):
    pass


def cmd_output_b(*cmd: str, **kwargs: Any) -> bytes:
    proc = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        **kwargs,
    )
    stdout, stderr = proc.communicate()

    if proc.returncode:
        raise CalledProcessError(cmd, proc.returncode, stdout, stderr)

    return stdout


def cmd_output(*cmd: str, **kwargs: Any) -> str:
    return cmd_output_b(*cmd, **kwargs).decode()



import argparse

from git_code_debt.generate_config import DEFAULT_GENERATE_CONFIG_FILENAME


def add_color(argparser: argparse.ArgumentParser) -> None:
    argparser.add_argument(
        '--color',
        default='auto',
        choices=('always', 'never', 'auto'),
        help='Whether to use color in output.',
    )


def add_generate_config_filename(argparser: argparse.ArgumentParser) -> None:
    argparser.add_argument(
        '-C', '--config-filename', default=DEFAULT_GENERATE_CONFIG_FILENAME,
        help='Path to generate config.',
    )



import datetime
import json

import flask
from werkzeug.wrappers import Response

from git_code_debt.server import logic
from git_code_debt.server.render_mako import render_template
from git_code_debt.util.time import data_points_for_time_range
from git_code_debt.util.time import to_timestamp


graph = flask.Blueprint('graph', __name__)


@graph.route('/graph/<metric_name>')
def show(metric_name: str) -> str:
    start_timestamp = int(flask.request.args['start'])
    end_timestamp = int(flask.request.args['end'])

    metric_info = logic.get_metric_info(flask.g.db, metric_name)

    data_points = data_points_for_time_range(
        start_timestamp,
        end_timestamp,
        data_points=250,
    )
    metrics_for_dates = logic.metrics_for_dates(metric_info.id, data_points)

    metrics_for_js = sorted({
        (m.date * 1000, m.value) for m in metrics_for_dates
    })

    return render_template(
        'graph.mako',
        description=metric_info.description,
        metric_name=metric_name,
        metrics=json.dumps(metrics_for_js),
        start_ts=start_timestamp,
        end_ts=end_timestamp,
        changes_url=flask.url_for(
            'changes.show',
            metric_name=metric_name,
            start_ts=start_timestamp,
            end_ts=end_timestamp,
        ),
    )


@graph.route('/graph/<metric_name>/all_data')
def all_data(metric_name: str) -> Response:
    earliest_timestamp = logic.get_first_data_timestamp(metric_name)
    now = datetime.datetime.today()

    return flask.redirect(
        flask.url_for(
            'graph.show',
            metric_name=metric_name,
            start=str(earliest_timestamp),
            end=str(to_timestamp(now)),
        ),
    )



import itertools
from typing import Generator
from typing import Iterable
from typing import Tuple
from typing import TypeVar

T = TypeVar('T')


def chunk_iter(
        iterable: Iterable[T],
        n: int,
) -> Generator[Tuple[T, ...], None, None]:
    """Yields an iterator in chunks

    For example you can do

    for a, b in chunk_iter([1, 2, 3, 4, 5, 6], 2):
        print('{} {}'.format(a, b))

    # Prints
    # 1 2
    # 3 4
    # 5 6

    Args:
        iterable - Some iterable
        n - Chunk size (must be greater than 0)
    """
    assert n > 0
    iterable = iter(iterable)

    chunk = tuple(itertools.islice(iterable, n))
    while chunk:
        yield chunk
        chunk = tuple(itertools.islice(iterable, n))



import re
from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Pattern

import cfgv


DEFAULT_GENERATE_CONFIG_FILENAME = 'generate_config.yaml'
SCHEMA = cfgv.Map(
    'Config', 'repo',

    cfgv.Required('repo', cfgv.check_string),
    cfgv.Required('database', cfgv.check_string),
    cfgv.Optional('skip_default_metrics', cfgv.check_bool, False),
    cfgv.Optional(
        'metric_package_names', cfgv.check_array(cfgv.check_string), [],
    ),
    cfgv.Optional('exclude', cfgv.check_regex, '^$'),
)


class GenerateOptions(NamedTuple):
    skip_default_metrics: bool
    metric_package_names: List[str]
    repo: str
    database: str
    exclude: Pattern[bytes]

    @classmethod
    def from_yaml(cls, dct: Dict[str, Any]) -> 'GenerateOptions':
        dct = cfgv.apply_defaults(cfgv.validate(dct, SCHEMA), SCHEMA)
        return cls(
            skip_default_metrics=dct['skip_default_metrics'],
            metric_package_names=dct['metric_package_names'],
            repo=dct['repo'],
            database=dct['database'],
            exclude=re.compile(dct['exclude'].encode()),
        )



import inspect
import pkgutil
from types import ModuleType
from typing import Any
from typing import Callable
from typing import Set
from typing import Type


def discover(
        package: ModuleType,
        cls_match_func: Callable[[Type[Any]], bool],
) -> Set[Type[Any]]:
    """Returns a set of classes in the directory matched by cls_match_func

    Args:
        path - A Python package
        cls_match_func - Function taking a class and returning true if the
            class is to be included in the output.
    """
    matched_classes = set()

    for _, module_name, _ in pkgutil.walk_packages(
            # https://github.com/python/mypy/issues/1422
            package.__path__,  # type: ignore
            prefix=package.__name__ + '.',
    ):
        module = __import__(module_name, fromlist=['__trash'], level=0)

        # Check all the classes in that module
        for _, imported_class in inspect.getmembers(module, inspect.isclass):
            # Don't include things that are only there due to a side-effect of
            # importing
            if imported_class.__module__ != module.__name__:
                continue

            if cls_match_func(imported_class):
                matched_classes.add(imported_class)

    return matched_classes



import collections
from typing import Dict
from typing import Generator
from typing import List
from typing import Tuple

from identify import identify

from git_code_debt.file_diff_stat import FileDiffStat
from git_code_debt.metric import Metric
from git_code_debt.metrics.base import DiffParserBase
from git_code_debt.metrics.base import MetricInfo
from git_code_debt.metrics.common import ALL_TAGS
from git_code_debt.metrics.common import UNKNOWN
from git_code_debt.repo_parser import Commit


class LinesOfCodeParser(DiffParserBase):
    """Counts lines of code in a repository, overall and by file types."""

    def get_metrics_from_stat(
            self,
            _: Commit,
            file_diff_stats: Tuple[FileDiffStat, ...],
    ) -> Generator[Metric, None, None]:
        total_lines = 0
        lines_by_file_type: Dict[str, int] = collections.defaultdict(int)

        for file_diff_stat in file_diff_stats:
            lines_changed = (
                len(file_diff_stat.lines_added) -
                len(file_diff_stat.lines_removed)
            )

            # Track total overall
            total_lines += lines_changed

            filename = file_diff_stat.filename.decode('UTF-8')
            tags = identify.tags_from_filename(filename) or {UNKNOWN}

            for tag in tags:
                lines_by_file_type[tag] += lines_changed

        # Yield overall metric and one per type of expected mapping types
        yield Metric('TotalLinesOfCode', total_lines)
        for tag, val in lines_by_file_type.items():
            if tag in ALL_TAGS and val:
                yield Metric(f'TotalLinesOfCode_{tag}', val)

    def get_metrics_info(self) -> List[MetricInfo]:
        metric_names = [f'TotalLinesOfCode_{tag}' for tag in ALL_TAGS]
        metric_names.append('TotalLinesOfCode')
        return [MetricInfo(metric_name) for metric_name in metric_names]



import datetime
from typing import Dict
from typing import FrozenSet
from typing import List
from typing import NamedTuple
from typing import Tuple

import flask

from git_code_debt.server import logic
from git_code_debt.server.metric_config import Config
from git_code_debt.server.presentation.delta import Delta
from git_code_debt.server.render_mako import render_template
from git_code_debt.util.time import to_timestamp


index = flask.Blueprint('index', __name__)

DATE_NAMES_TO_TIMEDELTAS = (
    ('Last Day', datetime.timedelta(days=1)),
    ('Last Week', datetime.timedelta(days=7)),
    ('Last Month', datetime.timedelta(days=30)),
    ('Last 3 Months', datetime.timedelta(days=90)),
    ('Last 6 Months', datetime.timedelta(days=180)),
    ('Last Year', datetime.timedelta(days=365)),
)


class Metric(NamedTuple):
    name: str
    color_override: bool
    current_value: int
    historic_deltas: Tuple[Delta, ...]
    all_data_url: str

    @property
    def classname(self) -> str:
        if self.color_override:
            return 'color-override'
        else:
            return ''

    @classmethod
    def from_data(
            cls,
            metric_name: str,
            today_timestamp: int,
            offsets: List[Tuple[str, int]],
            current_values: Dict[str, int],
            metric_data: Dict[str, Dict[str, int]],
            color_overrides: FrozenSet[str],
    ) -> 'Metric':
        return cls(
            metric_name,
            metric_name in color_overrides,
            current_values[metric_name],
            tuple(
                Delta(
                    flask.url_for(
                        'graph.show',
                        metric_name=metric_name,
                        start=str(timestamp),
                        end=str(today_timestamp),
                    ),
                    (
                        current_values[metric_name] -
                        metric_data[time_name][metric_name]
                    ),
                )
                for time_name, timestamp in offsets
            ),
            flask.url_for('graph.all_data', metric_name=metric_name),
        )


class Group(NamedTuple):
    name: str
    metrics: Tuple[Metric, ...]


def format_groups(
        config: Config,
        metric_names: List[str],
        today_timestamp: int,
        offsets: List[Tuple[str, int]],
        current_values: Dict[str, int],
        metric_data: Dict[str, Dict[str, int]],
) -> List[Group]:
    metrics = {
        metric_name: Metric.from_data(
            metric_name,
            today_timestamp,
            offsets,
            current_values,
            metric_data,
            config.color_overrides,
        )
        for metric_name in metric_names
    }

    defined_groups = [
        Group(
            group.name, tuple(
                metrics[metric_name]
                for metric_name in metric_names if group.contains(metric_name)
            ),
        )
        for group in config.groups
    ]

    uncategorized_group = Group(
        'Uncategorized',
        tuple(
            metrics[metric_name]
            for metric_name in metric_names if not any(
                group.contains(metric_name) for group in config.groups
            )
        ),
    )

    all_group = Group(
        'All', tuple(metrics[metric_name] for metric_name in metric_names),
    )

    all_groups = defined_groups + [uncategorized_group, all_group]
    return [group for group in all_groups if group.metrics]


@index.route('/')
def show() -> str:
    metric_names = logic.get_metric_ids(flask.g.db)
    today = datetime.datetime.today()
    today_timestamp = to_timestamp(today)
    offsets = [
        (time_name, to_timestamp(today - offset))
        for (time_name, offset) in DATE_NAMES_TO_TIMEDELTAS
    ]
    current_values = logic.get_metrics_for_sha(logic.get_latest_sha())
    metric_data = {
        time_name:
        logic.get_metrics_for_sha(logic.get_sha_for_date(timestamp))
        for (time_name, timestamp) in offsets
    }

    return render_template(
        'index.mako',
        groups=format_groups(
            flask.g.config,
            metric_names,
            today_timestamp,
            offsets,
            current_values,
            metric_data,
        ),
    )



import contextlib
import subprocess
import tempfile
from typing import Generator
from typing import List
from typing import NamedTuple
from typing import Optional

from git_code_debt.util.iter import chunk_iter
from git_code_debt.util.subprocess import cmd_output
from git_code_debt.util.subprocess import cmd_output_b


class Commit(NamedTuple):
    sha: str
    date: int


BLANK_COMMIT = Commit('0' * 40, 0)

COMMIT_FORMAT = '--format=%H%n%ct'


class RepoParser:

    def __init__(self, git_repo: str) -> None:
        self.git_repo = git_repo
        self.tempdir: Optional[str] = None

    @contextlib.contextmanager
    def repo_checked_out(self) -> Generator[None, None, None]:
        assert not self.tempdir
        with tempfile.TemporaryDirectory(suffix='temp-repo') as self.tempdir:
            try:
                subprocess.check_call((
                    'git', 'clone',
                    '--no-checkout', '--quiet', '--shared',
                    self.git_repo, self.tempdir,
                ))
                yield
            finally:
                self.tempdir = None

    def get_commit(self, sha: str) -> Commit:
        output = cmd_output(
            'git', 'show', COMMIT_FORMAT, sha, cwd=self.tempdir,
        )
        sha, date = output.splitlines()[:2]

        return Commit(sha, int(date))

    def get_commits(self, since_sha: Optional[str] = None) -> List[Commit]:
        """Returns a list of Commit objects.

        Args:
           since_sha - (optional) A sha to search from
        """
        assert self.tempdir

        cmd = ['git', 'log', '--first-parent', '--reverse', COMMIT_FORMAT]
        if since_sha:
            commits = [self.get_commit(since_sha)]
            cmd.append(f'{since_sha}..HEAD')
        else:
            commits = []
            cmd.append('HEAD')

        output = cmd_output(*cmd, cwd=self.tempdir)

        for sha, date in chunk_iter(output.splitlines(), 2):
            commits.append(Commit(sha, int(date)))

        return commits

    def get_original_commit(self, sha: str) -> bytes:
        assert self.tempdir
        return cmd_output_b('git', 'show', sha, cwd=self.tempdir)

    def get_commit_diff(self, previous_sha: str, sha: str) -> bytes:
        assert self.tempdir
        return cmd_output_b(
            'git', 'diff', previous_sha, sha, '--no-renames',
            cwd=self.tempdir,
        )



from git_code_debt.file_diff_stat import FileDiffStat
from git_code_debt.metrics.base import SimpleLineCounterBase


class TODOCount(SimpleLineCounterBase):
    def line_matches_metric(
            self,
            line: bytes,
            file_diff_stat: FileDiffStat,
    ) -> bool:
        return b'TODO' in line



from typing import Any

import mako.lookup
import pkg_resources


template_lookup = mako.lookup.TemplateLookup(
    directories=[
        pkg_resources.resource_filename('git_code_debt.server', 'templates'),
    ],
    default_filters=['html_escape'],
    imports=['from mako.filters import html_escape'],
)


def render_template(template_name: str, **env: Any) -> str:
    template = template_lookup.get_template(template_name)
    return template.render(**env)



import re
from typing import Any
from typing import Dict
from typing import FrozenSet
from typing import List
from typing import NamedTuple
from typing import Pattern
from typing import Tuple


class Group(NamedTuple):
    name: str
    metrics: FrozenSet[str]
    metric_expressions: Tuple[Pattern[str], ...]

    def contains(self, metric_name: str) -> bool:
        return (
            metric_name in self.metrics or
            any(expr.search(metric_name) for expr in self.metric_expressions)
        )

    @classmethod
    def from_yaml(
            cls,
            name: str,
            metrics: List[str],
            metric_expressions: List[str],
    ) -> 'Group':
        if not metrics and not metric_expressions:
            raise TypeError(
                'Group {} must define at least one of '
                '`metrics` or `metric_expressions`'.format(name),
            )
        return cls(
            name,
            frozenset(metrics),
            tuple(re.compile(expr) for expr in metric_expressions),
        )


def _get_groups_from_yaml(yaml: List[Dict[str, Any]]) -> Tuple[Group, ...]:
    # A group dict maps it's name to a dict containing metrics and
    # metric_expressions
    # Here's an example yaml:
    # [{'Bar': {'metrics': ['Foo', 'Bar'], 'metric_expressions': ['^Baz']}}]
    return tuple(
        Group.from_yaml(
            next(iter(group_dict.keys())),
            next(iter(group_dict.values())).get('metrics', []),
            next(iter(group_dict.values())).get('metric_expressions', []),
        )
        for group_dict in yaml
    )


def _get_commit_links_from_yaml(
        yaml: Dict[str, str],
) -> Tuple[Tuple[str, str], ...]:
    # The CommitLinks will look like
    # LinkName: 'link_value'
    # OtherLinkName: 'other_link_value'
    # Here we'll alphabetize these and return a tuple of (LinkName, link_value)
    return tuple(sorted(yaml.items()))


class Config(NamedTuple):
    color_overrides: FrozenSet[str]
    commit_links: Tuple[Tuple[str, str], ...]
    groups: Tuple[Group, ...]
    widget_metrics: List[str]

    @classmethod
    def from_data(cls, data: Dict[str, Any]) -> 'Config':
        return cls(
            color_overrides=frozenset(data['ColorOverrides']),
            commit_links=_get_commit_links_from_yaml(data['CommitLinks']),
            groups=_get_groups_from_yaml(data['Groups']),
            widget_metrics=data['WidgetMetrics'],
        )
